--------------------------- MODULE Witness ---------------------------
EXTENDS System_Spec

(***************************************************************************)
(* A ``witness'' execution is a stream of system states that results from  *)
(* an interleaving of agent steps such that every agent takes a step       *)
(* infinitely often. This module gives a formal definition of a witness    *)
(* execution and proves that such an execution exists, through a recursive *)
(* function definition.                                                    *)
(***************************************************************************)

Is_witness(H) ==
\*  /\ H \in Stream(GoodSystemState)
  /\ H[1] = [ st |-> initState, msgs |-> {}, t |-> 1 ]
  /\ \A n \in Natp : \E a \in Agent : executes(a, H[n], H[n+1])
  /\ \A a \in Agent : \A n \in Natp :
        \E m \in Natp : m >= n /\ executes(a, H[m], H[m+1]) 

(***************************************************************************)
(* The following definitions introduce a witness execution through a       *)
(* recursive function definition.                                          *)
(***************************************************************************)

(***************************************************************************)
(* Choose next system time such that some agent can make a ``good'' step.  *)
(* The parameter r denotes the previous system state.                      *)
(***************************************************************************)
nTime(r) ==
  CHOOSE t \in Time : \E a \in Agent : GoodNextTime(r,a,t)

(***************************************************************************)
(* Among all the agents that can make a ``good'' step, choose the one that *)
(* has the earliest deadline.                                              *)
(***************************************************************************)
nAgentP(r,t,c) ==
  /\ instr(r.st[c]).early <= t /\ t <= instr(r.st[c]).late
  /\ \A d \in Agent: 
        instr(r.st[d]).early <= t /\ t <= instr(r.st[d]).late
        => instr(r.st[c]).late <= instr(r.st[d]).late

nAgent(r,t) == CHOOSE c \in Agent : nAgentP(r,t,c)

LEMMA nTime_nAgent ==
  ASSUME NEW r \in GoodSystemState
  PROVE  LET nt == nTime(r)
             na == nAgent(r, nt)
         IN  /\ nt \in Time
             /\ na \in Agent
             /\ GoodNextTime(r, na, nt)
             /\ nAgentP(r, nt, na)
<1>. DEFINE nt == nTime(r)
            na == nAgent(r, nt)
<1>1. /\ nt \in Time
      /\ \E a \in Agent : GoodNextTime(r, a, nt)
  BY Zenon DEF GoodSystemState, nTime
<1>2. /\ r.t <= nt
      /\ \A a \in Agent : nt <= instr(r.st[a]).late
  BY <1>1 DEF GoodNextTime
<1>. DEFINE cands == { a \in Agent : /\ instr(r.st[a]).early <= nt
                                     /\ nt <= instr(r.st[a]).late }
            P(dl) == \E c \in cands : dl = instr(r.st[c]).late
<1>. HIDE DEF P
<1>3. \E dl \in Nat : P(dl)
  BY <1>1, instr_type, Zenon
     DEF GoodSystemState, SystemState, DatedInstruction, GoodNextTime, Time, P
<1>4. PICK dl \in Nat : /\ P(dl)
                        /\ \A d \in 0 .. dl-1 : ~ P(d)
  BY <1>3, SmallestNatural, Isa
<1>5. PICK c \in cands : instr(r.st[c]).late = dl
  BY <1>4 DEF P
<1>6. nAgentP(r, nt, c)
  <2>. SUFFICES ASSUME NEW d \in Agent,
                       instr(r.st[d]).early <= nt, 
                       nt <= instr(r.st[d]).late
                PROVE  instr(r.st[c]).late <= instr(r.st[d]).late
    BY DEF nAgentP  (* first conjunct holds by c \in cands *)
  <2>1. P(instr(r.st[d]).late)
    BY DEF P
  <2>2. instr(r.st[d]).late \in Nat
    BY instr_type DEF GoodSystemState, SystemState, DatedInstruction, GoodNextTime, Time
  <2>. QED  BY <2>1, <2>2, <1>4, <1>5
<1>7. na \in Agent /\ nAgentP(r, nt, na)
  BY <1>6 DEF nAgent
<1>8. GoodNextTime(r, na, nt)
  BY <1>2, <1>7 DEF nAgentP, GoodNextTime
<1>. QED  BY <1>1, <1>7, <1>8

(***************************************************************************)
(* Now define the witness as a recursive function. We introduce a separate *)
(* operator definition for the witness body so that we can refer to it in  *)
(* the following theorems.                                                 *)
(***************************************************************************)
rec_wit_body(w,n) ==
  IF n = 1 THEN [ st |-> initState, msgs |-> {}, t |-> 1 ]
  ELSE LET r == w[n-1]
           nt == nTime(r)
           na == nAgent(r,nt) 
           res == exec(r.st[na], r.msgs)
       IN  [ st |-> [r.st EXCEPT ![na] = res[1]], msgs |-> res[2], t |-> nt ]

rec_witness == CHOOSE w : w = [n \in Natp |-> rec_wit_body(w,n)]

LEMMA rec_wit_body_defined ==
  ASSUME NEW w, NEW n \in Natp,
         n>1 => w[n-1] \in GoodSystemState
  PROVE  rec_wit_body(w,n) \in GoodSystemState
<1>1. CASE n=1
  BY <1>1, initState_good DEF rec_wit_body
<1>2. CASE n>1
  <2>. DEFINE r == w[n-1]
              nt == nTime(r)
              na == nAgent(r,nt)
  <2>. r \in GoodSystemState
    BY <1>2
  <2>1. /\ nt \in Time
        /\ na \in Agent
        /\ nAgentP(r, nt, na)
        /\ GoodNextTime(r, na, nt)
    BY nTime_nAgent
  <2>2. rec_wit_body(w,n) \in SystemState
    BY <1>2, <2>1, exec_type 
       DEF Natp, GoodSystemState, SystemState, rec_wit_body
  <2>3. executes(na, r, rec_wit_body(w,n))
    BY <1>2, <2>1 DEF Natp, executes, rec_wit_body
  <2>. QED  BY <2>1, <2>2, <2>3, executes_preserves_GoodSystemState
<1>. QED  BY <1>1, <1>2 DEF Natp

LEMMA rec_witness_def ==
  /\ rec_witness = [n \in Natp |-> rec_wit_body(rec_witness,n)]
  /\ rec_witness \in Stream(GoodSystemState)
  /\ Is_witness(rec_witness)
<1>1. IsWellFoundedOn(OpToRel(<,Natp), Natp)
  BY Natp_wellfounded
<1>2. IsTransitivelyClosedOn(OpToRel(<,Natp), Natp)
  BY DEF IsTransitivelyClosedOn, OpToRel, Natp
<1>3. WFDefOn(OpToRel(<,Natp), Natp, rec_wit_body)
  <2> ASSUME NEW g, NEW h, NEW n \in Natp, \A i \in 1..n-1: g[i] = h[i]
      PROVE rec_wit_body(g,n) = rec_wit_body(h,n)
    <3>1. CASE n=1
      BY <3>1 DEF rec_wit_body
    <3>2. CASE n#1
      <4>. g[n-1] = h[n-1]
        BY <3>2 DEF Natp
      <4>. QED  BY <3>2 DEF rec_wit_body, nTime, nAgent, Natp
    <3>. QED  BY <3>1, <3>2
  <2> QED  BY DEF WFDefOn, SetLessThan, OpToRel, Natp
        
<1>4. OpDefinesFcn(rec_witness, Natp, rec_wit_body)
  BY DEF OpDefinesFcn, rec_witness
<1>5. rec_witness = [n \in Natp |-> rec_wit_body(rec_witness,n)]
  BY <1>1, <1>2, <1>3, <1>4, WFInductiveDefLemma, Isa DEF WFInductiveDefines

<1>6. rec_witness \in Stream(GoodSystemState)
  <2>1. GoodSystemState # {}
    BY initState_good
  <2>2. ASSUME NEW g \in [Natp -> GoodSystemState], NEW n \in Natp 
        PROVE  rec_wit_body(g, n) \in GoodSystemState
    BY rec_wit_body_defined DEF Natp
  <2>. QED
    BY <1>1, <1>3, <1>5, <2>1, <2>2, WFInductiveDefType, Natp_wellfounded, Isa
       DEF WFInductiveDefines, rec_witness, Stream

<1>7. Is_witness(rec_witness)
  <2> USE DEF Natp
  <2>2. rec_witness[1] = [ st |-> initState, msgs |-> {}, t |-> 1 ]
    BY <1>5 DEF rec_wit_body
  <2>3. ASSUME NEW n \in Natp
        PROVE  \E a \in Agent : executes(a, rec_witness[n], rec_witness[n+1])
     <3>. DEFINE r == rec_witness[n]
                 newT == nTime(r)
                 c == nAgent(r, newT)
                 res == exec(r.st[c], r.msgs)
     <3>1. rec_witness[n+1] = rec_wit_body(rec_witness, n+1)
       BY <1>5
     <3>2. /\ c \in Agent
           /\ GoodNextTime(r, c, newT)
       BY <1>6, nTime_nAgent DEF Stream
     <3>3. executes(c, r, rec_witness[n+1])
       BY <3>1, <3>2 DEF rec_wit_body, executes
     <3> QED  BY <3>2, <3>3
  <2>4. ASSUME NEW a \in Agent, NEW n \in Natp
        PROVE  \E m \in Natp : m >= n /\ executes(a, rec_witness[m], rec_witness[m+1]) 
    <3>. SUFFICES ASSUME \A m \in Natp : m >=n => ~executes(a, rec_witness[m], rec_witness[m+1])
                  PROVE  FALSE
      OBVIOUS
    <3>1. instr(rec_witness[n].st[a]).late \in Time
      BY <1>6, instr_type, Zenon DEF GoodSystemState, SystemState, DatedInstruction, Stream
    <3>2. \A k \in Nat : rec_witness[n+k].t \in Time
      BY <1>6 DEF GoodSystemState, SystemState, Stream
    <3>10. \A k \in Nat : rec_witness[n+k].st[a] = rec_witness[n].st[a]
      <4>. DEFINE P(k) == rec_witness[n+k].st[a] = rec_witness[n].st[a]
      <4>1. P(0)
        OBVIOUS
      <4>2. ASSUME NEW k \in Nat, P(k)
            PROVE  P(k+1)
        <5>1. PICK c \in Agent : executes(c, rec_witness[n+k], rec_witness[n+k+1])
          BY <2>3
        <5>2. c # a
          BY <5>1  (* and assumption that a doesn't execute beyond step n *)
        <5>3. rec_witness[n+k].st \in [Agent -> State]
          BY <1>6 DEF GoodSystemState, SystemState, Stream
        <5>4. rec_witness[n+k+1].st[a] = rec_witness[n+k].st[a]
          BY <5>1, <5>2, <5>3 DEF executes
        <5>. QED  BY <4>2, <5>4, Isa DEF Natp
      <4>3. QED   BY <4>1, <4>2, NatInduction, Isa
    <3>20. ASSUME NEW k \in Nat
           PROVE  rec_witness[n+k].t <= instr(rec_witness[n].st[a]).late
      <4>2. PICK t \in Time :
              /\ rec_witness[n+k].t <= t
              /\ t <= instr(rec_witness[n+k].st[a]).late
        BY <1>6 DEF GoodSystemState, GoodNextTime, Stream
      <4>3. instr(rec_witness[n+k].st[a]).late = instr(rec_witness[n].st[a]).late
        BY <3>10, Zenon
      <4>4. rec_witness[n+k].t \in Time
        BY <3>2, Zenon
      <4>. QED  BY ONLY <3>1, <4>4, <4>2, <4>3 DEF Time
    <3>30. PICK k \in Nat : rec_witness[n+k].t > instr(rec_witness[n].st[a]).late
      <4>1. PICK i \in Natp : rec_witness[i].t > instr(rec_witness[n].st[a]).late
        BY <1>6, <2>3, <3>1, TimeProgress, Zenon DEF GoodSystemState, Stream
      <4>2. ASSUME i <= n PROVE FALSE
        <5>. DEFINE rit == rec_witness[i].t
                    rnt == rec_witness[n].t
                    dl == instr(rec_witness[n].st[a]).late
        <5>1. /\ rit \in Time
              /\ rnt \in Time
              /\ dl \in Time
          BY <3>1, <1>6 DEF GoodSystemState, SystemState, Stream
        <5>2. rit <= rnt
          BY <1>6, <2>3, <4>2, executes_monotonic DEF GoodSystemState, Stream
        <5>3. rnt <= dl
          <6>1. PICK b \in Agent, t \in Time : GoodNextTime(rec_witness[n],b,t)
            BY <1>6 DEF GoodSystemState, Stream
          <6>2. /\ rnt <= t
                /\ t <= dl
            BY <6>1 DEF GoodNextTime 
          <6>. HIDE DEF rnt, dl
          <6>. QED  BY <6>2, <5>1 DEF Time
        <5>. HIDE DEF rit, rnt, dl
        <5>4. ~(dl < rit)
          BY <5>1, <5>2, <5>3 DEF Time
        <5>. QED  BY <4>1, <5>4, Zenon DEF rit, dl
      <4>3. i-n \in Nat
        BY <4>2
      <4>. WITNESS i-n \in Nat
      <4>. QED  BY <4>1
    <3>35. rec_witness[n+k].t <= instr(rec_witness[n].st[a]).late
      BY <3>20, Zenon
    <3>40. rec_witness[n+k].t \in Time
      BY <3>2, Zenon
    <3>. QED
      BY ONLY <3>35, <3>30, <3>1, <3>40 DEF Time
  <2>. QED  BY <2>2, <2>3, <2>4 DEF Is_witness, Stream
<1> QED  BY <1>5, <1>6, <1>7

LEMMA witness_existence == \E w \in Stream(GoodSystemState) : Is_witness(w)
BY rec_witness_def

LEMMA witness_filter_extend ==
  ASSUME NEW w \in Stream(SystemState), NEW a \in Agent, 
         NEW k \in Natp, NEW l \in Natp, l >= k,
         \A i \in k..l : w[i].st[a] = w[k].st[a]
  PROVE  filter(Restrict(w, 1..l), a) = filter(Restrict(w, 1..k), a)
<1>. DEFINE sk == Restrict(w, 1..k)
            sl == Restrict(w, 1..l)
            ska == project_agent(sk, a)
            sla == project_agent(sl, a)
            kla == SubSeq(sla, k+1, l)
<1>1. /\ sk \in Seq(SystemState)
      /\ sl \in Seq(SystemState)
      /\ Len(sk) = k
      /\ Len(sl) = l
  BY RestrictStream DEF Natp
<1>2. /\ ska \in Seq(State)
      /\ ska # << >>
      /\ sla \in Seq(State)
      /\ Len(ska) = k
      /\ Len(sla) = l
  BY <1>1, project_agent_type DEF Natp
<1>3. ska = SubSeq(sla, 1, k)
  <2>. /\ SubSeq(sla, 1, k) \in Seq(State)
       /\ Len(SubSeq(sla, 1, k)) = k
       /\ \A i \in 1..k : SubSeq(sla, 1, k)[i] = sla[i]
    BY <1>2, SubSeqProperties DEF Natp
  <2>. \A i \in 1 .. k : ska[i] = sla[i]
    BY DEF Stream, Restrict, Natp, project_agent
  <2>. QED  BY <1>2, SeqEqual, Zenon
<1>4. /\ kla \in Seq(State)
      /\ Len(kla) = l-k
      /\ \A i \in 1 .. l-(k+1)+1 : kla[i] = sla[(k+1)+i-1]
  <2>. HIDE DEF sla
  <2>. QED
    BY <1>2, k+1 \in 1 .. Len(sla)+1, l \in (k+1)-1 .. Len(sla), l - (k+1) + 1 = l-k,
       SubSeqProperties, Zenon DEF Natp
<1>5. sla = ska \o kla
  <2>. HIDE DEF ska, sla
  <2>. ska \o kla = SubSeq(sla, 1, l)
    BY <1>2, <1>3, 1 \in 1 .. Len(sla)+1, k \in 1-1 .. Len(sla), l \in k .. Len(sla),
       ConcatAdjacentSubSeq, Zenon DEF Natp
  <2>. QED  BY <1>2, SubSeqFull, Zenon
<1>6. \A i \in 1 .. Len(kla) : kla[i] = Last(ska)
  <2>1. \A i \in 1..l : sla[i] = w[i].st[a]
    BY DEF Restrict, project_agent, Natp
  <2>. HIDE DEF kla, sla
  <2>2. \A i \in 1 .. Len(kla) : kla[i] = w[k+i].st[a]
    BY <1>4, <2>1, Z3 DEF Natp
  <2>3. Last(ska) = w[k].st[a]
    BY <1>2 DEF Restrict, project_agent, Natp, Last
  <2>. QED  BY <1>4, <2>2, <2>3, Z3 DEF Natp
<1>7. unstutter[sla] = unstutter[ska]
  BY <1>2, <1>4, <1>5, <1>6, unstutter_concat_stuttered_last, Zenon
<1>. QED  BY <1>7 DEF filter


LEMMA witness_ordered ==
  ASSUME NEW w \in Stream(GoodSystemState), Is_witness(w),
         NEW i \in Natp, NEW j \in Natp, w[i].t < w[j].t
  PROVE  i < j
<1> USE DEF Time, Natp
<1> \A k \in Natp: w[k].t \in Time
  BY DEF GoodSystemState,SystemState, Stream
<1> SUFFICES ASSUME NEW u \in Natp
             PROVE  \A k \in Nat : w[u].t <= w[u+k].t
  BY Z3
<1> DEFINE P(k) == w[u].t <= w[u+k].t
<1>Base P(0)
  OBVIOUS
<1>Ind ASSUME NEW k \in Nat, P(k) PROVE P(k+1)
  <2>2 PICK a \in Agent: executes(a,w[u+k],w[u+k+1])
    BY DEF Is_witness                               
  <2>3 w[u+k].t <= w[u+k+1].t
    BY <2>2 DEF executes, GoodNextTime
  <2> QED  BY <2>3, <1>Ind DEF GoodSystemState, SystemState
<1> QED  BY <1>Base, <1>Ind, NatInduction, Isa
 

LEMMA witness_executes ==
  ASSUME NEW w \in Stream(GoodSystemState), Is_witness(w),
         NEW i \in Natp, NEW a \in Agent
  PROVE  \E j \in Natp : /\ j >= i
                         /\ \A k \in i..j : w[k].st[a] = w[i].st[a]
                         /\ executes(a, w[j], w[j+1])
<1> USE DEF Natp
<1> DEFINE P(j) == executes(a, w[i+j], w[i+j+1])
<1>1 PICK m \in Nat: P(m)
  <2>1 PICK k \in Natp: k >= i /\ executes(a,w[k],w[k+1])
    BY Zenon DEF Is_witness
  <2>. QED  BY <2>1
<1> HIDE DEF P
<1>2 PICK j \in Nat: P(j) /\ \A k \in 0..j-1: ~P(k)
  BY <1>1, SmallestNatural, Isa
<1> DEFINE Q(k) == k <= j => w[i+k].st[a] = w[i].st[a]
<1>Init Q(0)
  OBVIOUS
<1>Next ASSUME NEW k \in Nat, Q(k) PROVE Q(k+1)
  <2> HAVE k+1 <= j
  <2>1 PICK b \in Agent: executes(b,w[i+k],w[i+k+1])
    BY DEF Is_witness
  <2>2 a # b
    BY <1>2, <2>1, Z3 DEF P
  <2>3 w[i+(k+1)].st[a] = w[i+k].st[a]
    BY <2>1, <2>2 DEF executes, GoodSystemState, SystemState, Stream
  <2> QED  BY <1>Next, <2>3
<1>3 \A k \in Nat: Q(k)
  BY <1>Init, <1>Next, NatInduction, Isa
<1> QED  BY <1>2, <1>3, Z3 DEF P
      
LEMMA witness_monotonic_deadlines ==
  ASSUME NEW w \in Stream(GoodSystemState), Is_witness(w),
         NEW i \in Natp, NEW j \in Natp, i <= j,
         NEW a \in Agent
  PROVE  instr(w[i].st[a]).late <= instr(w[j].st[a]).late
<1> USE DEF Time, Natp
<1> /\ instr(w[i].st[a]).late \in Time
    /\ \A k \in Nat: instr(w[i+k].st[a]).late \in Time
  BY instr_type DEF GoodSystemState, SystemState, DatedInstruction, Stream
<1> DEFINE P(k) == instr(w[i].st[a]).late <= instr(w[i+k].st[a]).late
<1>Init P(0)
  OBVIOUS
<1>Next ASSUME NEW k \in Nat, P(k) PROVE P(k+1)
  <2>2 PICK b \in Agent: executes(b,w[i+k],w[i+k+1])
    BY DEF Is_witness
  <2>3 instr(w[i+k].st[a]).late <= instr(w[i+(k+1)].st[a]).late
    <3> USE DEF executes, GoodSystemState, SystemState, Stream
    <3>1 CASE a = b
      BY exec_monotonic, <2>2, <3>1
    <3>2 CASE a # b
      BY <2>2, <3>2, w[i+(k+1)].st[a] = w[i+k].st[a], instr_type, Isa DEF DatedInstruction
    <3> QED  BY <3>1, <3>2
  <2> QED  BY <1>Next, <2>3, Z3
<1> \A k \in Nat: P(k)
  BY <1>Init, <1>Next, NatInduction, Isa
<1> QED  BY Z3
  

LEMMA witness_msgs_monotonic == 
  ASSUME NEW w \in Stream(GoodSystemState), Is_witness(w),
         NEW i \in Natp, NEW j \in Nat, i <= j
  PROVE w[i].msgs \subseteq w[j].msgs
<1> DEFINE P(k) == w[i].msgs \subseteq w[i+k].msgs
<1> USE DEF Natp, Stream
<1>Init P(0)
  OBVIOUS
<1>Next ASSUME NEW k \in Nat, P(k) PROVE P(k+1)
  <2>2 PICK a \in Agent: executes(a,w[i+k],w[i+k+1])
    BY DEF Is_witness
  <2>3 w[i+k].msgs \subseteq w[i+(k+1)].msgs
    BY exec_msgs, <2>2 DEF executes, GoodSystemState, SystemState
  <2> QED  BY <1>Next, <2>3
<1> \A k \in Nat: P(k)
  BY <1>Init, <1>Next, NatInduction, Isa
<1> QED  BY Z3

(***************************************************************************)
(* The following lemma asserts that deadlines of send instructions cannot  *)
(* be missed. Otherwise that agent could never again execute, violating    *)
(* the liveness condition. The lemma could be generalized to arbitrary     *)
(* instructions.                                                           *)
(***************************************************************************)
LEMMA witness_sends_messages ==
  ASSUME NEW w \in Stream(GoodSystemState), Is_witness(w),
         NEW i \in Natp, NEW j \in Natp, NEW a \in Agent,
         instr(w[i].st[a]).ins.type = "send",
         instr(w[i].st[a]).late < w[j].t
  PROVE  instr(w[i].st[a]) \in w[j].msgs
<1> USE DEF Natp, Stream, Time
<1> instr(w[i].st[a]) \in DatedInstruction
  BY instr_type DEF GoodSystemState, SystemState
<1>1 PICK n \in Natp: /\ n >= i
                      /\ \A k \in i..n : w[k].st[a] = w[i].st[a]
                      /\ executes(a, w[n], w[n+1])
  BY witness_executes
<1>2 w[n].t < w[j].t
  <2> /\ w[n].t \in Time 
      /\ w[j].t \in Time
      /\ instr(w[i].st[a]).late \in Time
    BY DEF GoodSystemState, SystemState, DatedInstruction
  <2> QED  BY <1>1, GoodSystemState_late
<1>3 n < j
  BY <1>2, witness_ordered
<1>4 instr(w[i].st[a]) \in w[n+1].msgs
  BY <1>1 DEF executes, exec, DatedInstruction
<1>5 w[n+1].msgs \subseteq w[j].msgs
  BY <1>3, witness_msgs_monotonic
<1> QED  BY <1>4, <1>5

LEMMA witness_messages == 
  ASSUME NEW w \in Stream(SystemState), Is_witness(w), NEW i \in Natp
  PROVE  w[i].msgs \subseteq UNION { sent(Range(project_agent(Restrict(w,1..i),c))) : c \in Agent }
<1>. DEFINE rhs(n) == UNION { sent(Range(project_agent(Restrict(w,1..n-1),c))) : c \in Agent }
            P(n) == w[n].msgs \subseteq rhs(n)
<1>1. \A n \in Natp :
         /\ Restrict(w,1..n-1) \in Seq(SystemState)
         /\ Restrict(w,1..n) \in Seq(SystemState)
         /\ DOMAIN Restrict(w,1..n-1) = 1 .. n-1
         /\ DOMAIN Restrict(w,1..n) = 1 .. n
  BY RestrictStream, LenProperties DEF Natp
<1>2. \A c \in Agent, n \in Natp :
         /\ project_agent(Restrict(w,1..n-1),c) \in Seq(State)
         /\ project_agent(Restrict(w,1..n),c) \in Seq(State)
  BY <1>1, project_agent_type
<1>3. ASSUME NEW n \in Natp
      PROVE  rhs(n) \subseteq rhs(n+1)
 <2>1. IsPrefix(Restrict(w, 1..n-1), Restrict(w, 1..n))
   BY RestrictStreamMonotonic DEF Natp
 <2>2. ASSUME NEW c \in Agent
       PROVE  IsPrefix(project_agent(Restrict(w,1..n-1),c),
                       project_agent(Restrict(w,1..n),c))
   <3>. /\ Restrict(w,1..n-1) \in Seq(SystemState)
        /\ Restrict(w,1..n) \in Seq(SystemState)
     BY <1>1
   <3>. QED  BY <2>1, myIsPrefixMap, Isa DEF project_agent
 <2>3. \A c \in Agent : Range(project_agent(Restrict(w,1..n-1),c))
              \subseteq Range(project_agent(Restrict(w,1..n),c))
   BY <2>2, <1>2, myRangeIsPrefix
 <2>. QED  BY <2>3 DEF sent, Natp
<1>4. P(1)
  BY DEF Is_witness
<1>5. ASSUME NEW n \in Natp, P(n)
      PROVE  P(n+1)
  <2>1. PICK a \in Agent : executes(a, w[n], w[n+1])
    BY DEF Is_witness
  <2>. DEFINE di == instr(w[n].st[a])
  <2>2. di \in DatedInstruction
    BY instr_type DEF Stream, SystemState
  <2>3. w[n+1].msgs = exec(w[n].st[a], w[n].msgs)[2]
    BY <2>1 DEF executes
  <2>4. CASE di.ins.type = "send"
    <3>1. w[n+1].msgs = w[n].msgs \cup {di}
      BY <2>3, <2>4 DEF exec
    <3>2. rhs(n) \cup {di} \subseteq rhs(n+1)
      BY <1>3, <2>2, <2>4 DEF sent, project_agent, Message, Range, Restrict, Natp, Stream
    <3>. QED  BY <1>5, <3>1, <3>2
  <2>5. CASE di.ins.type # "send"
    BY <1>3, <1>5, <2>3, <2>5 DEF exec
  <2>. QED  BY <2>4, <2>5
<1>6. \A n \in Natp : P(n)
  BY <1>4, <1>5, Natp_induction, Isa
<1>. QED  BY <1>3, <1>6 DEF Natp

=============================================================================